/*
 * @(#)ImagePanel.java	1.0 08/15/04
 *
 * McGill Univarsity
 */

import java.awt.*;
import javax.swing.*;
import java.awt.image.*;
import javax.imageio.ImageIO;
import java.awt.geom.AffineTransform;
import java.io.File;

/**
 * A <code>JPanel</code> that displays, loads, saves and modifies images.
 *
 * The <code>refreshImage</code> method displays an image based on the contents
 * of an image file. The <code>saveImage</code> method saves the displayed image
 * to file. The <code>rotate</code> method rotates the displayed image. The
 * <code>removeHorizontalLines</code> method removes horizontal lines in the
 * displayed image based on user set thresholds concerning colour and vertical
 * run length. The <code>showSelectedPixels</code> method highlights certain pixels
 * that meet certain user set thresholds concerning colour and vertical run length,
 * and is useful for debugging. The <code>paintComponent</code> method  is needed
 * in order to display images and changes to them.
 *
 * @author Cory McKay
 */ 
public class ImagePanel 
	extends JPanel
{
	/* FIELDS ******************************************************************/

	
	private	MainFrame		parent_window;	// the window that instantiated this panel
	private BufferedImage	image;			// the image that is displayed and processed
	private int				image_width;	// the widthh of image
	private int				image_height;	// the height of image


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Loads the image file with the given path into this object's
	 * image field and displays it.
	 *
	 * @param	file_path		The path of the file to be loaded.
	 * @param	parent_window	The window that instantiated this panel.
	 */
	public ImagePanel (String file_path, MainFrame parent_window)
	{
		this.parent_window = parent_window;
		refreshImage(file_path);
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Loads the image file with the specified path, stores it in this object's
	 * image field and displays it. Automatically displays an error message and
	 * opens a file called <i>example.jpg</i> if no valid image file exists that
	 * matches the specified file path. Displays an error message and quits the
	 * program if this file cannot be loaded.
	 *
	 * <p>The width and height of the image are stored in this classes fields.
	 *
	 * @param	file_path	The path of the file to be loaded.
	 */
	public void refreshImage(String file_path)
	{
		// Load the image as an Image and then store it in the image BufferedImage field.
		// Perform proper error handling.
		try
		{
			Image img = getToolkit().createImage(file_path);
			MediaTracker t = new MediaTracker(this);
			t.addImage(img, 0);
			t.waitForAll();
			image = new BufferedImage(img.getWidth(this), img.getHeight(this),
				BufferedImage.TYPE_INT_RGB);
			image.createGraphics().drawImage(img, 0, 0, this);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, "Could not open image file specified.\nOpening example.jpg instead.", "ERROR", JOptionPane.ERROR_MESSAGE);
			Image img = getToolkit().createImage("example.jpg");
			MediaTracker t = new MediaTracker(this);
			t.addImage(img, 0);
			try
			{
				t.waitForAll();
				image = new BufferedImage(img.getWidth(this), img.getHeight(this),
					BufferedImage.TYPE_INT_RGB);
				image.createGraphics().drawImage(img, 0, 0, this);
				file_path = "example.jpg";
			}
			catch (Exception f)
			{
				JOptionPane.showMessageDialog(null, "Could not open default example image file specified.\nQuitting application.", "ERROR", JOptionPane.ERROR_MESSAGE);
				System.exit(0);
			}
		}

		// Store image width and size
		image_width = image.getWidth();
		image_height = image.getHeight();
		setPreferredSize (new Dimension(image_width, image_height));

		// Update the file path in the parent_window and its text display
		parent_window.setNewFilePath(file_path);

		// Refresh the display of the image
		repaint();
	}

	
	/**
	 * Saves the image that is currently held in this object's field to the
	 * specified file. It is assumed that the calling method has ensured
	 * that this file can be written to. The image is saved in jpeg format.
	 *
	 * @param	file_to_save_to	The file to save the image to.
	 * @throws	Exception		Throws an exception if there is a problem saving the file.
	 */
	public void saveImage(File file_to_save_to)
		throws Exception
	{
		// Save the image field to the file_to_save_to file
		ImageIO.write(image, "JPG", file_to_save_to);
		
		// Update the file path in the parent_window and its text display
		parent_window.setNewFilePath(file_to_save_to.getPath());
	}


	/**
	 * Rotates the image that is currently held in this object's field by
	 * the amount specified by the <code>getRotationAmount</code> method
	 * of the <i>rotate_dialog_box</i> paramater.
	 *
	 * @param	rotate_dialog_box	Contains the angle by which to rotate the image.
	 */
	public void rotate(RotateDialogBox rotate_dialog_box)
	{
		double rotation_degrees = rotate_dialog_box.getRotationAmount();
		double rotation_radians = Math.toRadians(rotation_degrees);
		AffineTransform transform = AffineTransform.getRotateInstance( rotation_radians,
		                                                               image_width / 2,
		                                                               image_height / 2 );
		AffineTransformOp op = new AffineTransformOp(transform, AffineTransformOp.TYPE_BILINEAR);
		BufferedImage rotated_image = new BufferedImage(image_width, image_height, image.getType());
		op.filter(image, rotated_image);
		image = rotated_image;
		repaint();
	}


	/**
	 * Removes horizontal lines from the image held in this object's field.
	 * The pixels that are removed must fall within certain colour and/or
	 * run length thresholds. These thresholds are specified in the 
	 * <i>horizontal_line_removal_dialog_box</i> parameter.
	 *
	 * <p>Pixels that are selected through the above method as belonging to
	 * a horizontal line are altered by applying a blur filter that averages
	 * their new colour to that of the immediately surrounding pixels that 
	 * are not selected.
	 *
	 * <p>Displays an error message if a problem occurs.
	 *
	 * @param	horizontal_line_removal_dialog_box	Specifies parameters needed to determine which pixels belong to horizontal lines.
	 */
	public void removeHorizontalLines(HorizontalLineRemovalSettingsDialogBox horizontal_line_removal_dialog_box)
	{
		try
		{
			// Find the pixels that are considered horizontal lines
			boolean[][] selected_pixels = (getMapsPassingSelectedThresholds(horizontal_line_removal_dialog_box))[5];

			// Apply blur filter to change colour of selected pixels
			meldSelectedPixels(selected_pixels);

			// Display results
			repaint();
		}
		catch (Exception e)
		{
			// Display an error message if a problem occured.
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
		}
	}


	/**
	 * Highlights pixels in the image held in this object's field that pass the 
	 * succesive filter requirements of the filters selected and there related
	 * thresholds in the given <i>horizontal_line_removal_dialog_box</i>
	 * parameter.
	 *
	 * <p>Only the results of the last <i>n</i> filter layers, as set by the 
	 * <i>Max Layers To Display</i> text area of the
	 * <i>horizontal_line_removal_dialog_box</i> are displayed. The pixels
	 * surviving the last stage of filtering are set to bright green, those surviving the
	 * second to last stage are set to bright orange, those surviving the third to last stage 
	 * are set to blue, those survivng the fouth to last stage are set to magenta, and those
	 * survivng the first stage are set to cyan.
	 *
	 * <p>Displays an error message if a problem occurs.
	 *
	 * @param	horizontal_line_removal_dialog_box	Specifies parameters needed to determine which pixels belong to horizontal lines.
	 */
	public void showSelectedPixels(HorizontalLineRemovalSettingsDialogBox horizontal_line_removal_dialog_box)
	{
		try
		{
			// Throw informative exception if maximum layers to display is not between 1 and 5
			if ( horizontal_line_removal_dialog_box.getLayersToDisplay() < 1 ||
				 horizontal_line_removal_dialog_box.getLayersToDisplay() > 5 )
				throw new Exception( "Max layers to display is " + horizontal_line_removal_dialog_box.getLayersToDisplay() + ".\n" +
									 "It should be between 1 and 5." );
			
			// Find the survivors of each filtering stage
			boolean[][][] pixel_maps = getMapsPassingSelectedThresholds(horizontal_line_removal_dialog_box);

			// The colours to use for highlighting pixels
			Color[] colours = new Color[5];
			colours[0] = Color.cyan;
			colours[1] = Color.magenta;
			colours[2] = Color.blue;
			colours[3] = Color.orange;
			colours[4] = Color.green;

			// Find how many filters were used
			int filters_used = 0;
			for (int i = 0; i < pixel_maps.length - 1; i++)
				if (pixel_maps[i] != null)
					filters_used++;

			// Display the pixels selcected by the appropriate filters
			int max_number_filter_results_to_display = horizontal_line_removal_dialog_box.getLayersToDisplay();
			int ok_to_display_filter = max_number_filter_results_to_display - filters_used;
			int current_colour_indice = colours.length - max_number_filter_results_to_display;
			if (max_number_filter_results_to_display > filters_used)
				current_colour_indice = max_number_filter_results_to_display - filters_used;
			for (int i = 0; i < (pixel_maps.length - 1); i++)
				if (pixel_maps[i] != null)
				{
					if (ok_to_display_filter >= 0)
					{
						colourPixels(pixel_maps[i], colours[current_colour_indice]);
						current_colour_indice++;
					}
					else
						ok_to_display_filter++;
				}

			// Display results
			repaint();
		}
		catch (Exception e)
		{
			// Display an error message if a problem occured.
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
		}
	}

	
	/**
	 * Allow the image and changes to the image to be displayed. Not used
	 * directly by this software.
	 */
	public void paintComponent (Graphics g)
	{
		super.paintComponent (g);
		g.drawImage (image, 0, 0, null);
	}


	/* PRIVATE METHODS *********************************************************/


	/**
	 * Finds maps for which pixels in the <i>image</i> field survive each of the
	 * filtering stages selected in the given <i>horizontal_line_removal_dialog_box</i>
	 * parameter.
	 *
	 * <p>Only those filters selected for use in <i>horizontal_line_removal_dialog_box</i>
	 * are used, and the threshold values in <i>horizontal_line_removal_dialog_box</i>
	 * are used. The filters are applied succesively in the following order:
	 *
	 * <p>- Absolute RGB Channel Value Threshold
	 * <br>- Relative RGB Channel Value Threshold
	 * <br>- Brightness Threshold
	 * <br>- Maximum Vertical Run Length
	 * <br>- Minimum Horizontal Run Length
	 *
	 * <p>These filters are applied succesively in a way such that only those pixels
	 * surviving the previous filtering stage are elligible to survive the next filter.
	 * All pixels are initially candidates.
	 *
	 * <p>The returned array contains six pixel maps, in the following order:
	 *
	 * <p>- Survivors of Absolute RGB Channel Value Threshold
	 * <br>- Survivors of Relative RGB Channel Value Threshold and previous filters
	 * <br>- Survivors of Brightness Threshold and previous filters
	 * <br>- Survivors of Vertical Run Length and previous filters
	 * <br>- Survivors of Horizontal Run Length and previous filters
	 * <br>- Survivors of all filters
	 *
	 * <p>Each of these is a 2-D array whose first indice represents the x-coordinate of
	 * a pixel and whose second indice represents the y-coordinate of a pixel.
	 *
	 * <p>Any of the first five  maps will be null if the corresponding option was not
	 * selected.
	 *
	 * <p>Displays an error message if a problem occurs.
	 *
	 * @param	horizontal_line_removal_dialog_box	Specifies parameters needed to determine which pixels survive each filter and which filters are applied.
	 * @throw	Exception							Informative exception if a problem occurs.
	 */
	private boolean[][][] getMapsPassingSelectedThresholds(HorizontalLineRemovalSettingsDialogBox horizontal_line_removal_dialog_box)
		throws Exception
	{
		// Prepare the map of candidate pixels before filtering
		boolean[][] candidate_pixels = new boolean[image_width][image_height];
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				candidate_pixels[x][y] = true;

		// Determine which pixels fall within the absolute RGB channel thresholds
		boolean[][] pixels_passing_absolute_RGB_values = null;
		if ( horizontal_line_removal_dialog_box.getFiltersToUse()[0] )
		{
			pixels_passing_absolute_RGB_values = findAbsoluteChannelValuePixels( horizontal_line_removal_dialog_box.getLowRedThreshold(),
																				 horizontal_line_removal_dialog_box.getHighRedThreshold(),
																				 horizontal_line_removal_dialog_box.getLowGreenThreshold(),
																				 horizontal_line_removal_dialog_box.getHighGreenThreshold(),
																				 horizontal_line_removal_dialog_box.getLowBlueThreshold(),
																				 horizontal_line_removal_dialog_box.getHighBlueThreshold(),
																				 candidate_pixels );
			candidate_pixels = pixels_passing_absolute_RGB_values;
		}

		// Determine which pixels fall within the relative RGB channel thresholds
		boolean[][] pixels_passing_relative_RGB_values = null;
		if ( horizontal_line_removal_dialog_box.getFiltersToUse()[1] )
		{
			pixels_passing_relative_RGB_values = findRelativeChannelValuePixels( horizontal_line_removal_dialog_box.getMinRedPercentageThreshold(),
																				 horizontal_line_removal_dialog_box.getMaxRedPercentageThreshold(),
																				 horizontal_line_removal_dialog_box.getMinGreenPercentageThreshold(),
																				 horizontal_line_removal_dialog_box.getMaxGreenPercentageThreshold(),
																				 horizontal_line_removal_dialog_box.getMinBluePercentageThreshold(),
																				 horizontal_line_removal_dialog_box.getMaxBluePercentageThreshold(),
																				 candidate_pixels );
			candidate_pixels = pixels_passing_relative_RGB_values;
		}

		// Determine which pixels fall within the brightness thresholds
		boolean[][] pixels_passing_brightness_threshold = null;
		if ( horizontal_line_removal_dialog_box.getFiltersToUse()[2] )
		{
			pixels_passing_brightness_threshold = findBrightnessPixels( horizontal_line_removal_dialog_box.getMinBrightnessThreshold(),
																		horizontal_line_removal_dialog_box.getMaxBrightnessThreshold(),
																	    candidate_pixels );
			candidate_pixels = pixels_passing_brightness_threshold;
		}

		// Determine which pixels have a sufficiently small vertical run length
		boolean[][] pixels_passing_vertical_run_length = null;
		if ( horizontal_line_removal_dialog_box.getFiltersToUse()[3] )
		{
			pixels_passing_vertical_run_length = findVerticalRunLengthMap( candidate_pixels,
																		   horizontal_line_removal_dialog_box.getVerticalRunLengthThreshold() );
			candidate_pixels = pixels_passing_vertical_run_length;
		}

		// Determine which pixels have a sufficiently large horizontal run length
		boolean[][] pixels_passing_horizontal_run_length = null;
		if ( horizontal_line_removal_dialog_box.getFiltersToUse()[4] )
		{
			pixels_passing_horizontal_run_length = findHorizontalRunLengthMap( candidate_pixels,
																			   horizontal_line_removal_dialog_box.getHorizontalRunLengthThreshold() );
			candidate_pixels = pixels_passing_horizontal_run_length;
		}

		// Prepare the results for the return format and return them
		boolean[][][] results = new boolean[6][][];
		results[0] = pixels_passing_absolute_RGB_values;
		results[1] = pixels_passing_relative_RGB_values;
		results[2] = pixels_passing_brightness_threshold;
		results[3] = pixels_passing_vertical_run_length;
		results[4] = pixels_passing_horizontal_run_length;
		results[5] = candidate_pixels;
		return results;
	}


	/**
	 * Finds the pixels in the <i>image</i> field that have absolute RGB colour channel
	 * values falling between the thresholds in the given parameters. These thresholds
	 * must be between 0 and 255, and the min threshold for each colour must not be higher
	 * than the corresponding max threshold. An informative exception is thrown if these
	 * requirements are not met in the parameters.
	 *
	 * Only pixels with (x, y) coordinates with corresponding entries of true in the
	 * given <i>candidate_pixel_map</i> may be candidates.
	 *
	 * The array returned has a first indice corresponding to horizontal pixel coordinate
	 * in <i>image</i>, and a second indice corresponding to vertical coordinate. Each entry
	 * is true if the corresponding pixel falls within all of the colour thresholds, and
	 * false if it does not.
	 */
	private boolean[][] findAbsoluteChannelValuePixels( int low_red_threshold, int high_red_threshold, 
	                                                    int low_green_threshold, int high_green_threshold,
	                                                    int low_blue_threshold, int high_blue_threshold,
	                                                    boolean[][] candidate_pixel_map )
		throws Exception
	{
		// Ensure that parameters are valid
		if (low_red_threshold > high_red_threshold)
			throw new Exception("The low absolute threshold for red detection must be lower than the high threshold.");
		if (low_green_threshold > high_green_threshold)
			throw new Exception("The low absolute threshold for green detection must be lower than the high threshold.");
		if (low_blue_threshold > high_blue_threshold)
			throw new Exception("The low absolute threshold for blue detection must be lower than the high threshold.");
		if (low_red_threshold < 0 || high_red_threshold > 255)
			throw new Exception( "Absolute thresholds for detecting red pixels must be between 0 and 255.\n" +
			                     "Low and high thresholds are respectively at " + low_red_threshold + " and " +
			                     high_red_threshold + ".");
		if (low_green_threshold < 0 || high_green_threshold > 255)
			throw new Exception( "Absolute thresholds for detecting green pixels must be between 0 and 255.\n" +
			                     "Low and high thresholds are respectively at " + low_green_threshold + " and " +
			                     high_green_threshold + ".");
		if (low_blue_threshold < 0 || high_blue_threshold > 255)
			throw new Exception( "Absolute thresholds for detecting blue pixels must be between 0 and 255.\n" +
			                     "Low and high thresholds are respectively at " + low_blue_threshold + " and " +
			                     high_blue_threshold + ".");

		// Get information about pixel colours of image
		ColorModel colour_model = image.getColorModel();
		WritableRaster raster = image.getRaster();

		// Prepare array to return
		boolean[][] selected_pixels = new boolean[image_width][image_height];
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				selected_pixels[x][y] = false;

		// Find the pixels within the colour thresholds
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				if (candidate_pixel_map[x][y])
				{
					int[] rgb = findPixelColourBrightnesses(x, y, colour_model, raster);
					int red_brightness = rgb[0];
					int green_brightness = rgb[1];
					int blue_brightness = rgb[2];
					if ( red_brightness >= low_red_threshold && red_brightness <= high_red_threshold &&
						 green_brightness >= low_green_threshold && green_brightness <= high_green_threshold &&
						 blue_brightness >= low_blue_threshold && blue_brightness <= high_blue_threshold )
						selected_pixels[x][y] = true;
				}

		// Return the results
		return selected_pixels;
	}


	/**
	 * Finds the pixels in the <i>image</i> field that have relative RGB colour channel
	 * values falling between the thresholds in the given parameters. These thresholds
	 * must be between 0% and 100%, and the min threshold for each colour must not be higher
	 * than the corresponding max threshold. An informative exception is thrown if these
	 * requirements are not met in the parameters.
	 *
	 * Only pixels with (x, y) coordinates with corresponding entries of true in the
	 * given <i>candidate_pixel_map</i> may be candidates.
	 *
	 * The array returned has a first indice corresponding to horizontal pixel coordinate
	 * in <i>image</i>, and a second indice corresponding to vertical coordinate. Each entry
	 * is true if the corresponding pixel falls within all of the colour thresholds, and
	 * false if it does not.
	 */
	private boolean[][] findRelativeChannelValuePixels( double low_red_threshold, double high_red_threshold, 
	                                                    double low_green_threshold, double high_green_threshold,
	                                                    double low_blue_threshold, double high_blue_threshold,
	                                                    boolean[][] candidate_pixel_map )
		throws Exception
	{
		// Ensure that parameters are valid
		if (low_red_threshold > high_red_threshold)
			throw new Exception("The low relative threshold for red detection must be lower than the high threshold.");
		if (low_green_threshold > high_green_threshold)
			throw new Exception("The low relative threshold for green detection must be lower than the high threshold.");
		if (low_blue_threshold > high_blue_threshold)
			throw new Exception("The low relative threshold for blue detection must be lower than the high threshold.");
		if (low_red_threshold < 0 || high_red_threshold > 100)
			throw new Exception( "Relative thresholds for detecting red pixels must be between 0% and 100%.\n" +
			                     "Low and high thresholds are respectively at " + low_red_threshold + " and " +
			                     high_red_threshold + ".");
		if (low_green_threshold < 0 || high_green_threshold > 100)
			throw new Exception( "Relative thresholds for detecting green pixels must be between 0% and 100%.\n" +
			                     "Low and high thresholds are respectively at " + low_green_threshold + " and " +
			                     high_green_threshold + ".");
		if (low_blue_threshold < 0 || high_blue_threshold > 100)
			throw new Exception( "Relative thresholds for detecting blue pixels must be between 0% and 100%.\n" +
			                     "Low and high thresholds are respectively at " + low_blue_threshold + " and " +
			                     high_blue_threshold + ".");
		
		ColorModel colour_model = image.getColorModel();
		WritableRaster raster = image.getRaster();

		// Prepare array to return
		boolean[][] selected_pixels = new boolean[image_width][image_height];
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				selected_pixels[x][y] = false;

		// Find the pixels within the colour thresholds
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				if (candidate_pixel_map[x][y])
				{
					int[] rgb = findPixelColourBrightnesses(x, y, colour_model, raster);
					double red_brightness = (double) rgb[0];
					double green_brightness = (double) rgb[1];
					double blue_brightness = (double) rgb[2];
					double total_brightness = red_brightness + green_brightness + blue_brightness;
					double red_percentage = (100.0 * red_brightness) / total_brightness;
					double green_percentage = (100.0 * green_brightness) / total_brightness;
					double blue_percentage = (100.0 * blue_brightness) / total_brightness;
					if ( red_percentage >= low_red_threshold && red_percentage <= high_red_threshold &&
						 green_percentage >= low_green_threshold && green_percentage <= high_green_threshold &&
						 blue_percentage >= low_blue_threshold && blue_percentage <= high_blue_threshold )
						selected_pixels[x][y] = true;
				}

		// Return the results
		return selected_pixels;
	}


	/**
	 * Finds the pixels in the <i>image</i> field that have an average brightness
	 * between the thresholds provided in the first two parameters. The brightness
	 * is determined by averaging the values of the three RGB channels. These thresholds
	 * must be between 0 and 255, and the min threshold must not be higher than the 
	 * corresponding max threshold. An informative exception is thrown if these
	 * requirements are not met in the parameters.
	 *
	 * Only pixels with (x, y) coordinates with corresponding entries of true in the
	 * given <i>candidate_pixel_map</i> may be candidates.
	 *
	 * The array returned has a first indice corresponding to horizontal pixel coordinate
	 * in <i>image</i>, and a second indice corresponding to vertical coordinate. Each entry
	 * is true if the corresponding pixel falls within the brightness thresholds, and
	 * false if it does not.
	 */
	private boolean[][] findBrightnessPixels( double low_brightness_threshold,
	                                          double high_brightness_threshold, 
	                                          boolean[][] candidate_pixel_map )
		throws Exception
	{
		// Ensure that parameters are valid
		if (low_brightness_threshold > high_brightness_threshold)
			throw new Exception("The low brightness threshold must be lower than the high threshold.");
		if (low_brightness_threshold < 0.0 || high_brightness_threshold > 255.0)
			throw new Exception( "Brightness thresholds must be between 0.0 and 255.0.\n" +
			                     "Low and high thresholds are respectively at " + low_brightness_threshold + " and " +
			                     high_brightness_threshold + ".");

		// Get information about pixel colours of image
		ColorModel colour_model = image.getColorModel();
		WritableRaster raster = image.getRaster();

		// Prepare array to return
		boolean[][] selected_pixels = new boolean[image_width][image_height];
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				selected_pixels[x][y] = false;

		// Find the pixels within the colour thresholds
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				if (candidate_pixel_map[x][y])
				{
					int[] rgb = findPixelColourBrightnesses(x, y, colour_model, raster);
					double red_brightness = (double) rgb[0];
					double green_brightness = (double) rgb[1];
					double blue_brightness = (double) rgb[2];
					double sum_of_brightnesses = red_brightness + green_brightness + blue_brightness;
					double average_brightness = (sum_of_brightnesses / 3.0);
					if ( average_brightness >= low_brightness_threshold && 
					     average_brightness <= high_brightness_threshold )
						selected_pixels[x][y] = true;
				}

		// Return the results
		return selected_pixels;
	}


	/**
	 * Finds the pixels in the <i>image</i> field that have corresponding entries in the
	 * <i>map</i> parameter of true (first indice is x coordingate and second is y
	 * coordinate) and also have vertical run lengths less than the value specified in the
	 * <i>run_length_threshold</i> parameter. An informative exception is thrown if the
	 * <i>run_length_threshold</i> parameter is less than 0.
	 *
	 * The array returned has a first indice corresponding to horizontal pixel coordinate
	 * in <i>image</i>, and a second indice corresponding to vertical coordinate. Each entry
	 * is true if the corresponding pixel meets the above conditions, and false if it does not.
	 */
	private boolean[][] findVerticalRunLengthMap(boolean[][] map, int run_length_threshold)
		throws Exception
	{
		// Throw informative exception if run_length_threshold is below 0
		if (run_length_threshold < 0)
			throw new Exception("Vertical run length threshold is set to " + run_length_threshold + ".\n" +
			                    "Vertical run length threshold must be positive.");
		
		// Prepare the array to return
		boolean[][] reduced_map = new boolean[image_width][image_height];
		for (int x = 0; x < image_width; x++)
			for (int y = 0; y < image_height; y++)
			 reduced_map[x][y] = false;
		
		// Find the pixels meeting the necessary conditions
		for (int x = 0; x < image_width; x++)
			for (int y = 0; y < image_height; y++)
			{
				int count = 0;
				if (map[x][y])
				{
					for (int y_2 = y; y_2 < image_height; y_2++)
					{
						if (map[x][y_2])
							count++;
						if ( !map[x][y_2] || y_2 == (image_height - 1) )
						{
							if (count <= run_length_threshold)
							{
								for (int y_3 = y; y_3 < y_2; y_3++)
									reduced_map[x][y_3] = true;
								if ( y_2 == (image_height - 1) && map[x][y_2] )
									reduced_map[x][y_2] = true;
							}
							count = 0;
							y = y_2; // continue from end of run
							y_2 = image_height + 1; // exit loop
						}
					}
				}
			}
	
		// Return the results
		return reduced_map;
	}


	/**
	 * Finds the pixels in the <i>image</i> field that have corresponding entries in the
	 * <i>map</i> parameter of true (first indice is x coordingate and second is y
	 * coordinate) and also have horizontal run lengths greater than the value specified in the
	 * <i>run_length_threshold</i> parameter. An informative exception is thrown if the
	 * <i>run_length_threshold</i> parameter is less than 0.
	 *
	 * The array returned has a first indice corresponding to horizontal pixel coordinate
	 * in <i>image</i>, and a second indice corresponding to vertical coordinate. Each entry
	 * is true if the corresponding pixel meets the above conditions, and false if it does not.
	 */
	private boolean[][] findHorizontalRunLengthMap(boolean[][] map, int run_length_threshold)
		throws Exception
	{
		// Throw informative exception if run_length_threshold is below 0
		if (run_length_threshold < 0)
			throw new Exception("Vertical run length threshold is set to " + run_length_threshold + ".\n" +
			                    "Vertical run length threshold must be positive.");
		
		// Prepare the array to return
		boolean[][] reduced_map = new boolean[image_width][image_height];
		for (int x = 0; x < image_width; x++)
			for (int y = 0; y < image_height; y++)
			 reduced_map[x][y] = false;
		
		// Find the pixels meeting the necessary conditions
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
			{
				int count = 0;
				if (map[x][y])
				{
					for (int x_2 = x; x_2 < image_width; x_2++)
					{
						if (map[x_2][y])
							count++;
						if ( !map[x_2][y] || x_2 == (image_width - 1) )
						{
							if (count >= run_length_threshold)
							{
								for (int x_3 = x; x_3 < x_2; x_3++)
									reduced_map[x_3][y] = true;
								if ( x_2 == (image_width - 1) && map[x_2][y] )
									reduced_map[x_2][y] = true;
							}
							count = 0;
							x = x_2; // continue from end of run
							x_2 = image_width + 1; // exit loop
						}
					}
				}
			}

		// Return the results
		return reduced_map;
	}


	/**
	 * Takes the pixels in the <i>image</i> field that have corresponding entries in the
	 * <i>map</i> parameter of true (first indice is x coordinate and second is y
	 * coordinate) and applies a blur filter that averages their new colour to that
	 * of the immediately surrounding pixels that do not have corresponding entries
	 * of true in the <i>map</i> parameter.
	 *
	 * <p><b>IMPORTANT:</b> It should be noted that this method alters the contents of
	 * the <i>map</i> parameter.
	 */
	private void meldSelectedPixels(boolean[][] map)
	{
		// Get inormation about the image field
		ColorModel colour_model = image.getColorModel();
		WritableRaster raster = image.getRaster();

		// Change colours of appropriate pixels
		boolean still_remaining = true; // whether or not there are still pixels that need to be altered
		while (still_remaining)
		{
			still_remaining = false;
			for (int x = 0; x < image_width; x++)
			{
				for (int y = 0; y < image_height; y++)
				{
					if (map[x][y])
					{
						// Set up array to hold surrounding colours
						Color[] surrounding_colours = new Color[8];
						for (int i = 0; i < surrounding_colours.length; i++)
							surrounding_colours[i] = null;

						// Find colours of surrounding pixels that are not selected in map
						if (y != 0)
							if (!map[x][y - 1])
								surrounding_colours[0] = getColour(x, y - 1, colour_model, raster);
						if (y != image_height - 1)
							if (!map[x][y + 1])
								surrounding_colours[1] = getColour(x, y + 1, colour_model, raster);
						if (x != 0)
							if (!map[x - 1][y])
								surrounding_colours[2] = getColour(x - 1, y, colour_model, raster);
						if (x != image_width - 1)
							if (!map[x + 1][y])
								surrounding_colours[3] = getColour(x + 1, y, colour_model, raster);
						if (y != 0 && x != 0)
							if (!map[x - 1][y - 1])
								surrounding_colours[4] = getColour(x - 1, y - 1, colour_model, raster);
						if (y != 0 && x != image_width - 1)
							if (!map[x + 1][y - 1])
								surrounding_colours[5] = getColour(x + 1, y - 1, colour_model, raster);
						if (y != image_height - 1 && x != 0)
							if (!map[x - 1][y + 1])
								surrounding_colours[6] = getColour(x - 1, y + 1, colour_model, raster);
						if (y != image_height - 1 && x != image_width - 1)
							if (!map[x + 1][y + 1])
								surrounding_colours[7] = getColour(x + 1, y + 1, colour_model, raster);

						// Find the total RGB brightnesses of the surrounding elligible pixels
						int count = 0; // number of surrounding elligible pixels
						int[] colour_averages = new int[3]; // averages of the surrounding elligible RGB values
						for (int i = 0; i < colour_averages.length; i++)
							colour_averages[i] = 0;
						for (int i = 0; i < surrounding_colours.length; i++)
							if (surrounding_colours[i] != null)
							{
								count++;
								colour_averages[0] += surrounding_colours[i].getRed();
								colour_averages[1] += surrounding_colours[i].getGreen();
								colour_averages[2] += surrounding_colours[i].getBlue();
							}

						// Change the colour of the pixel at the (x, y) coordinate
						if (count != 0)
						{
							// Find the average RGB colours
							for (int i = 0; i < colour_averages.length; i++)
								colour_averages[i] = colour_averages[i] / count;

							// Set the pixel to the new colour
							Color new_colour = new Color(colour_averages[0], colour_averages[1], colour_averages[2]);
							int argbOC = new_colour.getRGB();
							Object colorData = colour_model.getDataElements(argbOC, null);
							raster.setDataElements(x, y, colorData);

							// Mark the pixel as no longer selected on the map
							map[x][y] = false;
						}

						// Note that there are still some pixels whose colours could not be changed
						// because all of the surrounding pixels were still selected. This will
						// be fixed in future iterations.
						else
							still_remaining = true;
					}
				}
			}
		}
	}


	/**
	 * Returns the respective brightnesses for red, green and blue of the pixel at
	 * the (x, y) coordinate of of the given <code>WritableRaster</code> and according
	 * to the given <code>ColorModel</code>.
	 */
	private int[] findPixelColourBrightnesses(int x, int y, ColorModel model, WritableRaster raster)
	{
		Color colour = getColour(x, y, model, raster);

		int[] rgb = new int[3];
		rgb[0] = colour.getRed();
		rgb[1] = colour.getGreen();
		rgb[2] = colour.getBlue();

		return rgb;
	}

	
	/**
	 * Returns the <code>Color</code>  of the pixel at the (x, y) coordinate of of
	 * the given <code>WritableRaster</code> and according to the given 
	 * <code>ColorModel</code>.
	 */
	private Color getColour(int x, int y, ColorModel model, WritableRaster raster)
	{
		Object data = raster.getDataElements(x, y, null);
		int argb = model.getRGB(data);
		return new Color(argb, true);
	}


	/**
	 * Colours the pixels in the <i>image</i> field whose corresponding entries in
	 * the <i>map</i> parameter are marked marked true. The pixels are marked with
	 * the given <code>Color</code>.
	 */
	private void colourPixels(boolean[][] map, Color marking_colour)
	{
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		int argbOC = marking_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < image_height; y++)
			for (int x = 0; x < image_width; x++)
				if (map[x][y])
					raster.setDataElements(x, y, colorData);
	}
}